#!/usr/bin/python
#
# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import re

from google.datacatalog_connectors.commons import cleanup, ingest

from google.datacatalog_connectors.qlik import prepare, scrape
from google.datacatalog_connectors.qlik.prepare import constants


class MetadataSynchronizer:
    __ENTRY_GROUP_ID = 'qlik'
    __SPECIFIED_SYSTEM = 'qlik'
    __TAG_TEMPLATE_NAME_PATTERN = r'^(.+?)/tagTemplates/(?P<id>.+?)$'

    def __init__(self, qlik_server_address, qlik_ad_domain, qlik_username,
                 qlik_password, datacatalog_project_id,
                 datacatalog_location_id):

        self.__project_id = datacatalog_project_id
        self.__location_id = datacatalog_location_id

        self.__metadata_scraper = scrape.MetadataScraper(
            qlik_server_address, qlik_ad_domain, qlik_username, qlik_password)

        self.__tag_template_factory = prepare.DataCatalogTagTemplateFactory(
            project_id=datacatalog_project_id,
            location_id=datacatalog_location_id)

        self.__site_url = qlik_server_address

        self.__assembled_entry_factory = prepare.AssembledEntryFactory(
            project_id=datacatalog_project_id,
            location_id=datacatalog_location_id,
            entry_group_id=self.__ENTRY_GROUP_ID,
            user_specified_system=self.__SPECIFIED_SYSTEM,
            site_url=self.__site_url)

    def run(self):
        """Coordinates a full scrape > prepare > ingest process."""

        # Scrape metadata from the Qlik server.
        logging.info('')
        logging.info('===> Scraping Qlik Sense metadata...')

        logging.info('')
        logging.info('Objects to be scraped: Custom Property Definitions')
        custom_property_defs = self.__scrape_custom_property_definitions()

        logging.info('')
        logging.info('Objects to be scraped:'
                     ' Streams, Apps, Dimensions, Measures, and Sheets')
        streams = self.__scrape_streams()
        logging.info('==== DONE ========================================')

        # Prepare: convert Qlik metadata into Data Catalog entities model.
        logging.info('')
        logging.info('===> Converting Qlik metadata'
                     ' into Data Catalog entities model...')

        tag_templates_dict = self.__make_tag_templates_dict(
            custom_property_defs)

        assembled_entries_dict = self.__make_assembled_entries_dict(
            custom_property_defs, streams, tag_templates_dict)
        logging.info('==== DONE ========================================')

        # Data Catalog entry relationships mapping.
        logging.info('')
        logging.info('===> Mapping Data Catalog entry relationships...')

        self.__map_datacatalog_relationships(assembled_entries_dict)
        logging.info('==== DONE ========================================')

        # Data Catalog clean up: delete obsolete data.
        logging.info('')
        logging.info('===> Deleting Data Catalog obsolete metadata...')

        self.__delete_obsolete_entries(assembled_entries_dict)
        logging.info('==== DONE ========================================')

        # Ingest metadata into Data Catalog.
        logging.info('')
        logging.info('===> Synchronizing Qlik :: Data Catalog metadata...')

        self.__ingest_metadata(assembled_entries_dict, tag_templates_dict)
        logging.info('==== DONE ========================================')

    def __scrape_custom_property_definitions(self):
        """Scrapes metadata from all the Custom Property Definitions the
        current user has access to.

        :return: A ``list`` of Custom Property Definition metadata.
        """
        return self.__metadata_scraper.scrape_all_custom_property_definitions()

    def __scrape_streams(self):
        """Scrapes metadata from all the Streams the current user has access
        to.

        The returned metadata include nested objects such as Apps and
        Sheets.

        :return: A ``list`` of Stream metadata.
        """
        all_streams = self.__metadata_scraper.scrape_all_streams()
        all_apps = self.__metadata_scraper.scrape_all_apps()

        # Not being published means the app is a work in progress, so it can be
        # skipped.
        published_apps = [app for app in all_apps if app.get('published')]

        for app in published_apps:
            # The below fields are not available in the scrape apps API
            # response, so they are injected into the returned metadata object
            # to turn further processing more efficient.
            app['dimensions'] = self.__scrape_dimensions(app)
            app['measures'] = self.__scrape_measures(app)
            app['visualizations'] = self.__scrape_visualizations(app)
            app['sheets'] = self.__scrape_published_sheets(app)

        self.__assemble_streams_metadata_from_flat_lists(
            all_streams, published_apps)

        return all_streams

    def __scrape_dimensions(self, app):
        """Scrapes metadata from the Dimensions the current user has access to
        within the given App.

        :return: A ``list`` of dimension metadata.
        """
        dimensions = self.__metadata_scraper.scrape_dimensions(app)
        # The 'app' field is not available in the scrape dimensions API
        # response, so it is injected into the returned metadata object to turn
        # further processing more efficient.
        for dimension in dimensions:
            dimension['app'] = {
                'id': app.get('id'),
                'name': app.get('name'),
            }

        return dimensions

    def __scrape_measures(self, app):
        """Scrapes metadata from the Measures the current user has access to
        within the given App.

        :return: A ``list`` of measure metadata.
        """
        measures = self.__metadata_scraper.scrape_measures(app)
        # The 'app' field is not available in the scrape measures API
        # response, so it is injected into the returned metadata object to turn
        # further processing more efficient.
        for measure in measures:
            measure['app'] = {
                'id': app.get('id'),
                'name': app.get('name'),
            }

        return measures

    def __scrape_visualizations(self, app):
        """Scrapes metadata from the Visualizations the current user has access
        to within the given App.

        :return: A ``list`` of visualization metadata.
        """
        visualizations = self.__metadata_scraper.scrape_visualizations(app)
        # The 'app' field is not available in the scrape visualizations API
        # response, so it is injected into the returned metadata object to turn
        # further processing more efficient.
        for visualization in visualizations:
            visualization['app'] = {
                'id': app.get('id'),
                'name': app.get('name'),
            }

        return visualizations

    def __scrape_published_sheets(self, app):
        """Scrapes metadata from all the published Sheets the current user has
        access to within the given App.

        :return: A ``list`` of sheet metadata.
        """
        sheets = self.__metadata_scraper.scrape_sheets(app)
        # Not being published means the sheet is a work in progress, so it can
        # be skipped.
        published_sheets = [
            sheet for sheet in sheets if sheet.get('qMeta').get('published')
        ]
        # The 'app' field is not available in the scrape sheets API response,
        # so it is injected into the returned metadata object to turn further
        # processing more efficient.
        for sheet in published_sheets:
            sheet['app'] = {
                'id': app.get('id'),
                'name': app.get('name'),
            }

        return published_sheets

    @classmethod
    def __assemble_streams_metadata_from_flat_lists(cls, all_streams,
                                                    published_apps):
        """Assembles the streams metadata from the given flat asset lists.

        """
        streams_dict = {}
        # Build an id-based dict to promote faster lookup in the next steps.
        for stream in all_streams:
            streams_dict[stream.get('id')] = stream

        for app in published_apps:
            stream_id = app.get('stream').get('id')

            # The 'apps' field is not available in the scrape streams API
            # response, so it is injected into the returned metadata object to
            # turn further processing more efficient.
            if not streams_dict[stream_id].get('apps'):
                streams_dict[stream_id]['apps'] = []

            streams_dict[stream_id]['apps'].append(app)

        return all_streams

    def __make_tag_templates_dict(self, custom_property_defs):
        templates_dict = {}

        self.__add_template_to_dict(
            templates_dict,
            self.__tag_template_factory.make_tag_template_for_app())
        self.__add_template_to_dict(
            templates_dict,
            self.__tag_template_factory.
            make_tag_template_for_custom_property_definition())
        self.__add_template_to_dict(
            templates_dict,
            self.__tag_template_factory.make_tag_template_for_dimension())
        self.__add_template_to_dict(
            templates_dict,
            self.__tag_template_factory.make_tag_template_for_measure())
        self.__add_template_to_dict(
            templates_dict,
            self.__tag_template_factory.make_tag_template_for_visualization())
        self.__add_template_to_dict(
            templates_dict,
            self.__tag_template_factory.make_tag_template_for_sheet())
        self.__add_template_to_dict(
            templates_dict,
            self.__tag_template_factory.make_tag_template_for_stream())

        for definition in custom_property_defs:
            for value in definition.get('choiceValues') or []:
                self.__add_template_to_dict(
                    templates_dict,
                    self.__tag_template_factory.
                    make_tag_template_for_custom_property_value(
                        definition, value))

        return templates_dict

    def __add_template_to_dict(self, templates_dict, template):
        template_id = re.match(pattern=self.__TAG_TEMPLATE_NAME_PATTERN,
                               string=template.name).group('id')
        templates_dict[template_id] = template

    def __make_assembled_entries_dict(self, custom_property_defs_metadata,
                                      streams_metadata, tag_templates_dict):
        """Makes Data Catalog entries and tags for the Qlik assets the current
        user has access to.

        Returns:
            A ``dict`` in which keys are the top level asset ids and values are
            flat lists containing those assets and their nested ones, with all
            related entries and tags.
        """
        assembled_entries = {}

        property_def_tag_template = tag_templates_dict.get(
            constants.TAG_TEMPLATE_ID_CUSTOM_PROPERTY_DEFINITION)
        for property_def_metadata in custom_property_defs_metadata:
            assembled_entries[property_def_metadata.get('id')] = \
                [self.__assembled_entry_factory
                    .make_assembled_entry_for_custom_property_def(
                        property_def_metadata, property_def_tag_template)]

        for stream_metadata in streams_metadata:
            assembled_entries[stream_metadata.get('id')] = \
                self.__assembled_entry_factory\
                    .make_assembled_entries_for_stream(
                        stream_metadata, tag_templates_dict)

        return assembled_entries

    @classmethod
    def __map_datacatalog_relationships(cls, assembled_entries_dict):
        all_assembled_entries = []
        for assembled_entries_data in assembled_entries_dict.values():
            all_assembled_entries.extend(assembled_entries_data)

        prepare.EntryRelationshipMapper().fulfill_tag_fields(
            all_assembled_entries)

    def __delete_obsolete_entries(self, new_assembled_entries_dict):
        all_assembled_entries = []
        for assembled_entry_data in new_assembled_entries_dict.values():
            all_assembled_entries.extend(assembled_entry_data)

        cleanup.DataCatalogMetadataCleaner(
            self.__project_id, self.__location_id, self.__ENTRY_GROUP_ID). \
            delete_obsolete_metadata(
            all_assembled_entries,
            f'system={self.__SPECIFIED_SYSTEM}'
            f' tag:site_url:{self.__site_url}')

    def __ingest_metadata(self, assembled_entries_dict, tag_templates_dict):
        metadata_ingestor = ingest.DataCatalogMetadataIngestor(
            self.__project_id, self.__location_id, self.__ENTRY_GROUP_ID)

        entries_count = sum(
            len(entries) for entries in assembled_entries_dict.values())
        logging.info('==== %d entries to be synchronized!', entries_count)

        synced_entries_count = self.__ingest_custom_property_defs_metadata(
            assembled_entries_dict, tag_templates_dict, metadata_ingestor)
        synced_entries_count += self.__ingest_streams_metadata(
            assembled_entries_dict, tag_templates_dict, metadata_ingestor)

        logging.info('')
        logging.info('==== %d of %d entries successfully synchronized!',
                     synced_entries_count, entries_count)

    def __ingest_custom_property_defs_metadata(self, assembled_entries_dict,
                                               tag_templates_dict,
                                               metadata_ingestor):

        custom_property_defs_assembled_entries = []
        for assembled_entries in assembled_entries_dict.values():
            if assembled_entries[0].entry.user_specified_type == \
                    constants.USER_SPECIFIED_TYPE_CUSTOM_PROPERTY_DEFINITION:
                custom_property_defs_assembled_entries.extend(
                    assembled_entries)

        custom_property_defs_entries_count = \
            len(custom_property_defs_assembled_entries)

        logging.info('')
        logging.info(
            '==== %d Custom Property Definition entries to be ingested...',
            custom_property_defs_entries_count)

        if not custom_property_defs_assembled_entries:
            return 0

        required_templates_dict = self.__filter_required_tag_templates(
            custom_property_defs_assembled_entries, tag_templates_dict)
        metadata_ingestor.ingest_metadata(
            custom_property_defs_assembled_entries, required_templates_dict)

        return custom_property_defs_entries_count

    def __ingest_streams_metadata(self, assembled_entries_dict,
                                  tag_templates_dict, metadata_ingestor):

        stream_entries_dict = {}
        for stream_id, assembled_entries in assembled_entries_dict.items():
            if assembled_entries[0].entry.user_specified_type == \
                    constants.USER_SPECIFIED_TYPE_STREAM:
                stream_entries_dict[stream_id] = assembled_entries

        synced_entries_count = 0
        for stream_id, assembled_entries in stream_entries_dict.items():
            asset_entries_count = len(assembled_entries)

            logging.info('')
            logging.info(
                '==== The Stream identified by "%s" and its nested assets'
                ' comprise %d entries.', stream_id, asset_entries_count)

            required_templates_dict = self.__filter_required_tag_templates(
                assembled_entries, tag_templates_dict)
            metadata_ingestor.ingest_metadata(assembled_entries,
                                              required_templates_dict)
            synced_entries_count += asset_entries_count

        return synced_entries_count

    def __filter_required_tag_templates(self, assembled_entries,
                                        tag_templates_dict):
        """Filters the Tag Templates that are required to ingest the given
        Entries and their Tags.

        This utility method should be called before any call to
        DataCatalogMetadataIngestor.ingest_metadata(). The tag_templates_dict,
        containing a variety of Tag Templates created to support the whole sync
        workflow, is intended to be shared among several calls to
        ingest_metadata(). Each call ingests a subset of Entries and Tags, so
        the rationale here is to filter only the Tag Templates required for a
        specific metadata ingestion round. By doing this, it avoids unnecessary
        Tag Template processing, which includes billable API calls, during the
        ingestion process.

        :return: A ``dict``.
        """
        required_templates_dict = {}

        for assembled_entry in assembled_entries:
            for tag in assembled_entry.tags:
                template_id = re.match(
                    pattern=self.__TAG_TEMPLATE_NAME_PATTERN,
                    string=tag.template).group('id')
                required_templates_dict[template_id] = tag_templates_dict.get(
                    template_id)

        return required_templates_dict



#!/usr/bin/python
#
# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from google.cloud import datacatalog
from google.datacatalog_connectors.commons import prepare

from google.datacatalog_connectors.qlik.prepare import \
    constants, dynamic_properties_helper as dph


class DataCatalogTagTemplateFactory(prepare.BaseTagTemplateFactory):
    __ASCII_CHARACTER_ENCODING = 'ASCII'
    __BOOL_TYPE = datacatalog.FieldType.PrimitiveType.BOOL
    __DOUBLE_TYPE = datacatalog.FieldType.PrimitiveType.DOUBLE
    __STRING_TYPE = datacatalog.FieldType.PrimitiveType.STRING
    __TIMESTAMP_TYPE = datacatalog.FieldType.PrimitiveType.TIMESTAMP

    def __init__(self, project_id, location_id):
        self.__project_id = project_id
        self.__location_id = location_id

    def make_tag_template_for_app(self):
        tag_template = datacatalog.TagTemplate()

        tag_template.name = datacatalog.DataCatalogClient.tag_template_path(
            project=self.__project_id,
            location=self.__location_id,
            tag_template=constants.TAG_TEMPLATE_ID_APP)

        tag_template.display_name = 'Qlik App Metadata'

        self._add_primitive_type_field(tag_template, 'id', self.__STRING_TYPE,
                                       'Unique Id')

        self._add_primitive_type_field(tag_template, 'owner_username',
                                       self.__STRING_TYPE, 'Owner username')

        self._add_primitive_type_field(tag_template, 'owner_name',
                                       self.__STRING_TYPE, 'Owner name')

        self._add_primitive_type_field(tag_template, 'modified_by_username',
                                       self.__STRING_TYPE,
                                       'Username who modified it')

        self._add_primitive_type_field(tag_template, 'published',
                                       self.__BOOL_TYPE, 'Published')

        self._add_primitive_type_field(tag_template, 'publish_time',
                                       self.__TIMESTAMP_TYPE, 'Publish time')

        self._add_primitive_type_field(tag_template, 'last_reload_time',
                                       self.__TIMESTAMP_TYPE,
                                       'Last reload time')

        self._add_primitive_type_field(tag_template, 'stream_id',
                                       self.__STRING_TYPE, 'Stream Id')

        self._add_primitive_type_field(tag_template, 'stream_name',
                                       self.__STRING_TYPE, 'Stream name')

        self._add_primitive_type_field(tag_template, 'stream_entry',
                                       self.__STRING_TYPE,
                                       'Data Catalog Entry for the Stream')

        self._add_primitive_type_field(tag_template, 'file_size',
                                       self.__STRING_TYPE, 'File size')

        self._add_primitive_type_field(tag_template, 'thumbnail',
                                       self.__STRING_TYPE, 'Thumbnail')

        self._add_primitive_type_field(tag_template,
                                       'saved_in_product_version',
                                       self.__STRING_TYPE,
                                       'Saved in product version')

        self._add_primitive_type_field(tag_template, 'migration_hash',
                                       self.__STRING_TYPE, 'Migration hash')

        self._add_primitive_type_field(tag_template, 'availability_status',
                                       self.__DOUBLE_TYPE,
                                       'Availability status')

        self._add_primitive_type_field(tag_template, 'site_url',
                                       self.__STRING_TYPE,
                                       'Qlik Sense site url')

        return tag_template

    def make_tag_template_for_custom_property_definition(self):
        tag_template = datacatalog.TagTemplate()

        tag_template.name = datacatalog.DataCatalogClient.tag_template_path(
            project=self.__project_id,
            location=self.__location_id,
            tag_template=constants.TAG_TEMPLATE_ID_CUSTOM_PROPERTY_DEFINITION)

        tag_template.display_name = 'Qlik Custom Property Definition Metadata'

        self._add_primitive_type_field(tag_template, 'id', self.__STRING_TYPE,
                                       'Unique Id')

        self._add_primitive_type_field(tag_template, 'modified_by_username',
                                       self.__STRING_TYPE,
                                       'Username who modified it')

        self._add_primitive_type_field(tag_template, 'value_type',
                                       self.__STRING_TYPE, 'Value type')

        self._add_primitive_type_field(tag_template, 'choice_values',
                                       self.__STRING_TYPE, 'Choice values')

        self._add_primitive_type_field(tag_template, 'object_types',
                                       self.__STRING_TYPE, 'Object types')

        self._add_primitive_type_field(tag_template, 'site_url',
                                       self.__STRING_TYPE,
                                       'Qlik Sense site url')

        return tag_template

    def make_tag_template_for_custom_property_value(self, definition_metadata,
                                                    value):

        tag_template = datacatalog.TagTemplate()

        template_id = dph.DynamicPropertiesHelper\
            .make_id_for_custom_property_value_tag_template(
                definition_metadata, value)
        tag_template.name = datacatalog.DataCatalogClient.tag_template_path(
            project=self.__project_id,
            location=self.__location_id,
            tag_template=template_id)

        tag_template.display_name = dph.DynamicPropertiesHelper\
            .make_display_name_for_custom_property_value_tag_template(
                definition_metadata, value)

        self._add_primitive_type_field(tag_template, 'id', self.__STRING_TYPE,
                                       'Unique Id')

        self._add_primitive_type_field(tag_template, 'created_date',
                                       self.__TIMESTAMP_TYPE, 'Created date')

        self._add_primitive_type_field(tag_template, 'modified_date',
                                       self.__TIMESTAMP_TYPE, 'Modified date')

        self._add_primitive_type_field(tag_template, 'modified_by_username',
                                       self.__STRING_TYPE,
                                       'Username who modified it')

        self._add_primitive_type_field(tag_template, 'value',
                                       self.__STRING_TYPE, 'Value')

        self._add_primitive_type_field(tag_template, 'property_definition_id',
                                       self.__STRING_TYPE,
                                       'Property Definition Id')

        # According to the Qlik Analytics Platform Architecture Team, there was
        # no way of searching assets by the Custom Property values using Qlik
        # when this feature was implemented (Dec, 2020), which means the
        # catalog search might be helpful to address such a use case. Hence the
        # 'definition_' part was supressed from this Tag Field Id to turn seach
        # queries more intuitive, e.g. tag:property_name:<PROPERTY-NAME>.
        self._add_primitive_type_field(tag_template, 'property_name',
                                       self.__STRING_TYPE,
                                       'Property Definition name')

        self._add_primitive_type_field(
            tag_template, 'property_definition_entry', self.__STRING_TYPE,
            'Data Catalog Entry for the Property Definition')

        self._add_primitive_type_field(tag_template, 'site_url',
                                       self.__STRING_TYPE,
                                       'Qlik Sense site url')

        return tag_template

    def make_tag_template_for_dimension(self):
        tag_template = datacatalog.TagTemplate()

        tag_template.name = datacatalog.DataCatalogClient.tag_template_path(
            project=self.__project_id,
            location=self.__location_id,
            tag_template=constants.TAG_TEMPLATE_ID_DIMENSION)

        tag_template.display_name = 'Qlik Dimension Metadata'

        self._add_primitive_type_field(tag_template, 'id', self.__STRING_TYPE,
                                       'Unique Id')

        self._add_enum_type_field(tag_template, 'grouping', [
            constants.DIMENSION_GROUPING_SINGLE_TAG_FIELD,
            constants.DIMENSION_GROUPING_DRILL_DOWN_TAG_FIELD
        ], 'Grouping')

        self._add_primitive_type_field(tag_template, 'fields',
                                       self.__STRING_TYPE, 'Fields')

        self._add_primitive_type_field(tag_template, 'field_labels',
                                       self.__STRING_TYPE, 'Field labels')

        self._add_primitive_type_field(tag_template, 'tags',
                                       self.__STRING_TYPE, 'Qlik tags')

        self._add_primitive_type_field(tag_template, 'app_id',
                                       self.__STRING_TYPE, 'App Id')

        self._add_primitive_type_field(tag_template, 'app_name',
                                       self.__STRING_TYPE, 'App name')

        self._add_primitive_type_field(tag_template, 'app_entry',
                                       self.__STRING_TYPE,
                                       'Data Catalog Entry for the App')

        self._add_primitive_type_field(tag_template, 'site_url',
                                       self.__STRING_TYPE,
                                       'Qlik Sense site url')

        return tag_template

    def make_tag_template_for_measure(self):
        tag_template = datacatalog.TagTemplate()

        tag_template.name = datacatalog.DataCatalogClient.tag_template_path(
            project=self.__project_id,
            location=self.__location_id,
            tag_template=constants.TAG_TEMPLATE_ID_MEASURE)

        tag_template.display_name = 'Qlik Measure Metadata'

        self._add_primitive_type_field(tag_template, 'id', self.__STRING_TYPE,
                                       'Unique Id')

        self._add_primitive_type_field(tag_template, 'expression',
                                       self.__STRING_TYPE, 'Expression')

        self._add_primitive_type_field(tag_template, 'label_expression',
                                       self.__STRING_TYPE, 'Label expression')

        self._add_primitive_type_field(tag_template, 'is_custom_formatted',
                                       self.__BOOL_TYPE, 'Is custom formatted')

        self._add_primitive_type_field(tag_template, 'tags',
                                       self.__STRING_TYPE, 'Qlik tags')

        self._add_primitive_type_field(tag_template, 'app_id',
                                       self.__STRING_TYPE, 'App Id')

        self._add_primitive_type_field(tag_template, 'app_name',
                                       self.__STRING_TYPE, 'App name')

        self._add_primitive_type_field(tag_template, 'app_entry',
                                       self.__STRING_TYPE,
                                       'Data Catalog Entry for the App')

        self._add_primitive_type_field(tag_template, 'site_url',
                                       self.__STRING_TYPE,
                                       'Qlik Sense site url')

        return tag_template

    def make_tag_template_for_sheet(self):
        tag_template = datacatalog.TagTemplate()

        tag_template.name = datacatalog.DataCatalogClient.tag_template_path(
            project=self.__project_id,
            location=self.__location_id,
            tag_template=constants.TAG_TEMPLATE_ID_SHEET)

        tag_template.display_name = 'Qlik Sheet Metadata'

        self._add_primitive_type_field(tag_template, 'id', self.__STRING_TYPE,
                                       'Unique Id')

        self._add_primitive_type_field(tag_template, 'owner_username',
                                       self.__STRING_TYPE, 'Owner username')

        self._add_primitive_type_field(tag_template, 'owner_name',
                                       self.__STRING_TYPE, 'Owner name')

        self._add_primitive_type_field(tag_template, 'published',
                                       self.__BOOL_TYPE, 'Published')

        self._add_primitive_type_field(tag_template, 'publish_time',
                                       self.__TIMESTAMP_TYPE, 'Publish time')

        self._add_primitive_type_field(tag_template, 'approved',
                                       self.__BOOL_TYPE, 'Approved')

        self._add_primitive_type_field(tag_template, 'app_id',
                                       self.__STRING_TYPE, 'App Id')

        self._add_primitive_type_field(tag_template, 'app_name',
                                       self.__STRING_TYPE, 'App name')

        self._add_primitive_type_field(tag_template, 'app_entry',
                                       self.__STRING_TYPE,
                                       'Data Catalog Entry for the App')

        self._add_primitive_type_field(tag_template, 'source_object',
                                       self.__STRING_TYPE, 'Source object')

        self._add_primitive_type_field(tag_template, 'draft_object',
                                       self.__STRING_TYPE, 'Draft object')

        self._add_primitive_type_field(tag_template, 'site_url',
                                       self.__STRING_TYPE,
                                       'Qlik Sense site url')

        return tag_template

    def make_tag_template_for_stream(self):
        tag_template = datacatalog.TagTemplate()

        tag_template.name = datacatalog.DataCatalogClient.tag_template_path(
            project=self.__project_id,
            location=self.__location_id,
            tag_template=constants.TAG_TEMPLATE_ID_STREAM)

        tag_template.display_name = 'Qlik Stream Metadata'

        self._add_primitive_type_field(tag_template, 'id', self.__STRING_TYPE,
                                       'Unique Id')

        self._add_primitive_type_field(tag_template, 'owner_username',
                                       self.__STRING_TYPE, 'Owner username')

        self._add_primitive_type_field(tag_template, 'owner_name',
                                       self.__STRING_TYPE, 'Owner name')

        self._add_primitive_type_field(tag_template, 'modified_by_username',
                                       self.__STRING_TYPE,
                                       'Username who modified it')

        self._add_primitive_type_field(tag_template, 'site_url',
                                       self.__STRING_TYPE,
                                       'Qlik Sense site url')

        return tag_template

    def make_tag_template_for_visualization(self):
        tag_template = datacatalog.TagTemplate()

        tag_template.name = datacatalog.DataCatalogClient.tag_template_path(
            project=self.__project_id,
            location=self.__location_id,
            tag_template=constants.TAG_TEMPLATE_ID_VISUALIZATION)

        tag_template.display_name = 'Qlik Visualization Metadata'

        self._add_primitive_type_field(tag_template, 'id', self.__STRING_TYPE,
                                       'Unique Id')

        self._add_primitive_type_field(tag_template, 'title',
                                       self.__STRING_TYPE, 'Title')

        self._add_primitive_type_field(tag_template, 'subtitle',
                                       self.__STRING_TYPE, 'Subtitle')

        self._add_primitive_type_field(tag_template, 'footnote',
                                       self.__STRING_TYPE, 'Footnote')

        self._add_primitive_type_field(tag_template, 'type',
                                       self.__STRING_TYPE, 'Type')

        self._add_primitive_type_field(tag_template, 'tags',
                                       self.__STRING_TYPE, 'Qlik tags')

        self._add_primitive_type_field(tag_template, 'app_id',
                                       self.__STRING_TYPE, 'App Id')

        self._add_primitive_type_field(tag_template, 'app_name',
                                       self.__STRING_TYPE, 'App name')

        self._add_primitive_type_field(tag_template, 'app_entry',
                                       self.__STRING_TYPE,
                                       'Data Catalog Entry for the App')

        self._add_primitive_type_field(tag_template, 'site_url',
                                       self.__STRING_TYPE,
                                       'Qlik Sense site url')

        return tag_template



#!/usr/bin/python
#
# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import json
import logging

from urllib.parse import urlparse
import websockets

from google.datacatalog_connectors.qlik.scrape import \
    authenticator, constants, engine_api_dimensions_helper, \
    engine_api_measures_helper, engine_api_sheets_helper, \
    engine_api_visualizations_helper


class EngineAPIScraper:
    """Wraps requests to the Qlik Engine JSON API.

    The Qlik Engine JSON API is a WebSocket protocol that uses JSON to pass
    information between the Qlik associative engine and the clients. This API
    consists of a set of objects representing apps, lists, and so on. These
    objects are organized in a hierarchical structure.

    Websockets use an asynchronous communication channel, but the public
    methods from this class are intended to be called synchronously to keep
    consistency with the overall scrape > prepare > ingest  workflow. The
    public methods take care of handling the async API calls for their clients.

    Most private coroutines (async def) rely on 'async with' statements.
    They work with an asynchronous context manager and the connection is closed
    when exiting the context.

    Attributes:
        __auth_cookie: An HTTP cookie used to authorize the requests.
    """

    def __init__(self, server_address, ad_domain, username, password):
        self.__server_address = server_address
        self.__ad_domain = ad_domain
        self.__username = username
        self.__password = password

        # The server address starts with an http/https scheme. The below
        # statement replaces the original scheme with 'wss', which is used for
        # secure websockets communication.
        self.__base_api_endpoint = f'wss://{urlparse(server_address).hostname}'
        self.__common_headers = {
            constants.XRFKEY_HEADER_NAME: constants.XRFKEY,
        }

        self.__auth_cookie = None

    def get_dimensions(self, app_id):
        """Gets the Dimensions (Master Items) set up to a given App.

        Returns:
            A list of [GenericDimensionProperties](https://help.qlik.com/en-US/sense-developer/September2020/APIs/EngineAPI/definitions-GenericDimensionProperties.html).  # noqa E501
        """
        self.__set_up_auth_cookie()
        return engine_api_dimensions_helper.EngineAPIDimensionsHelper(
            self.__server_address, self.__auth_cookie).get_dimensions(app_id)

    def get_measures(self, app_id):
        """Gets the Measures (Master Items) set up to a given App.

        Returns:
            A list of [GenericMeasureProperties](https://help.qlik.com/en-US/sense-developer/September2020/APIs/EngineAPI/definitions-GenericMeasureProperties.html).  # noqa E501
        """
        self.__set_up_auth_cookie()
        return engine_api_measures_helper.EngineAPIMeasuresHelper(
            self.__server_address, self.__auth_cookie).get_measures(app_id)

    def get_sheets(self, app_id):
        """Gets the Sheets that belong to the given App.

        Returns:
            A list of [NxContainerEntry](https://help.qlik.com/en-US/sense-developer/September2020/APIs/EngineAPI/definitions-NxContainerEntry.html).  # noqa E501
        """
        self.__set_up_auth_cookie()
        return engine_api_sheets_helper.EngineAPISheetsHelper(
            self.__server_address, self.__auth_cookie).get_sheets(app_id)

    def get_visualizations(self, app_id):
        """Gets the Visualizations (Master Items) set up to a given App.

        Returns:
            A list of [GenericObjectProperties](https://help.qlik.com/en-US/sense-developer/September2020/APIs/EngineAPI/definitions-GenericObjectProperties.html).  # noqa E501
        """
        self.__set_up_auth_cookie()
        return engine_api_visualizations_helper.EngineAPIVisualizationsHelper(
            self.__server_address,
            self.__auth_cookie).get_visualizations(app_id)

    def __set_up_auth_cookie(self):
        if self.__auth_cookie:
            return

        windows_auth_url = asyncio.get_event_loop().run_until_complete(
            self.__get_windows_authentication_url())
        self.__auth_cookie = authenticator.Authenticator\
            .get_qps_session_cookie_windows_auth(
                ad_domain=self.__ad_domain,
                username=self.__username,
                password=self.__password,
                auth_url=windows_auth_url)
        logging.debug('QPS session cookie issued for the Engine API: %s',
                      self.__auth_cookie)

    async def __get_windows_authentication_url(self):
        """Gets a Windows Authentication url.

        This method sends an unauthenticated request to a well known endpoint
        of the Qlik Engine JSON API. The expected response has a `loginUri`
        param, which is the Windows Authentication url.
        P.S. The endpoint was manually captured from the Engine API Explorer's
        Execution Logs (https://<qlik-site>/dev-hub/engine-api-explorer).

        Returns:
            A string.
        """
        uri = f'{self.__base_api_endpoint}/app/?transient=' \
              f'?Xrfkey={constants.XRFKEY}' \
              f'&reloadUri={self.__server_address}/dev-hub/engine-api-explorer'

        # Sets the User-Agent to Windows temporarily to get a Windows
        # Authentication URL that is required by the NTLM authentication flow.
        headers = self.__common_headers.copy()
        headers['User-Agent'] = constants.WINDOWS_USER_AGENT

        async with websockets.connect(uri=uri,
                                      extra_headers=headers) as websocket:

            async for message in websocket:
                json_message = json.loads(message)
                params = json_message.get('params')
                if params:
                    return params.get('loginUri')



#!/usr/bin/python
#
# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import json
import logging

from google.datacatalog_connectors.qlik.scrape import \
    base_engine_api_helper, websocket_responses_manager


class EngineAPISheetsHelper(base_engine_api_helper.BaseEngineAPIHelper):
    # Methods to be used in the requests.
    __GET_OBJECTS = 'GetObjects'

    def get_sheets(self, app_id, timeout=60):
        try:
            return self._run_until_complete(self.__get_sheets(app_id, timeout))
        except Exception:
            logging.warning("error on get_sheets:", exc_info=True)
            return []

    async def __get_sheets(self, app_id, timeout):
        async with self._connect_websocket(app_id) as websocket:
            responses_manager = \
                websocket_responses_manager.WebsocketResponsesManager()

            await self._start_websocket_communication(websocket, app_id,
                                                      responses_manager)

            consumer = self.__consume_get_sheets_msg
            producer = self.__produce_get_sheets_msg
            return await asyncio.wait_for(
                self._handle_websocket_communication(
                    consumer(websocket, responses_manager),
                    producer(websocket, responses_manager)), timeout)

    async def __consume_get_sheets_msg(self, websocket, responses_manager):
        return await self._consume_messages(websocket, responses_manager,
                                            self.__GET_OBJECTS, 'result.qList')

    async def __produce_get_sheets_msg(self, websocket, responses_manager):
        return await self._produce_messages(
            websocket, responses_manager, self.__send_follow_up_msg_get_sheets)

    async def __send_follow_up_msg_get_sheets(self, websocket,
                                              responses_manager, response):

        response_id = response.get('id')
        if responses_manager.is_method(response_id, self._OPEN_DOC):
            await self.__handle_open_doc_response(websocket, responses_manager,
                                                  response)
            responses_manager.remove_unhandled(response)

    async def __handle_open_doc_response(self, websocket, responses_manager,
                                         response):

        doc_handle = response.get('result').get('qReturn').get('qHandle')
        follow_up_req_id = await self.__send_get_sheets_request(
            websocket, doc_handle)
        responses_manager.add_pending_id(follow_up_req_id, self.__GET_OBJECTS)

    async def __send_get_sheets_request(self, websocket, doc_handle):
        """Sends a Get Objects request for the sheet type.

        Returns:
            The request id.
        """
        request_id = self._generate_request_id()
        await websocket.send(
            json.dumps({
                'handle': doc_handle,
                'method': self.__GET_OBJECTS,
                'params': {
                    'qOptions': {
                        'qTypes': ['sheet'],
                    },
                },
                'id': request_id,
            }))

        logging.debug('Get Objects (type=sheet) request sent: %d', request_id)
        return request_id



#!/usr/bin/python
#
# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import json
import logging

from google.datacatalog_connectors.qlik.scrape import \
    base_engine_api_helper as base_helper, websocket_responses_manager


class EngineAPIVisualizationsHelper(base_helper.BaseEngineAPIHelper):
    # Keys used to identify the handles.
    __DOC_HANDLE = 'doc-handle'

    # Methods to be used in the requests.
    __GET_OBJECT = 'GetObject'
    __GET_PROPERTIES = 'GetProperties'

    def get_visualizations(self, app_id, timeout=60):
        try:
            return self._run_until_complete(
                self.__get_visualizations(app_id, timeout))
        except Exception:
            logging.warning("error on get_visualizations:", exc_info=True)
            return []

    async def __get_visualizations(self, app_id, timeout):
        async with self._connect_websocket(app_id) as websocket:
            responses_manager = \
                websocket_responses_manager.WebsocketResponsesManager()

            await self._start_websocket_communication(websocket, app_id,
                                                      responses_manager)

            consumer = self.__consume_get_visualizations_msg
            producer = self.__produce_get_visualizations_msg
            return await asyncio.wait_for(
                self._handle_websocket_communication(
                    consumer(websocket, responses_manager),
                    producer(websocket, responses_manager)), timeout)

    async def __consume_get_visualizations_msg(self, websocket,
                                               responses_manager):
        return await self._consume_messages(websocket, responses_manager,
                                            self.__GET_PROPERTIES,
                                            'result.qProp')

    async def __produce_get_visualizations_msg(self, websocket,
                                               responses_manager):
        return await self._produce_messages(
            websocket, responses_manager,
            self.__send_follow_up_msg_get_visualizations)

    async def __send_follow_up_msg_get_visualizations(self, websocket,
                                                      responses_manager,
                                                      response):

        response_id = response.get('id')
        if responses_manager.is_method(response_id, self._OPEN_DOC):
            await self.__handle_open_doc_response(websocket, responses_manager,
                                                  response)
            responses_manager.remove_unhandled(response)
        elif responses_manager.is_method(response_id, self._GET_ALL_INFOS):
            await self.__handle_get_all_infos_response(websocket,
                                                       responses_manager,
                                                       response)
            responses_manager.remove_unhandled(response)
        elif responses_manager.is_method(response_id, self.__GET_OBJECT):
            await self.__handle_get_object_response(websocket,
                                                    responses_manager,
                                                    response)
            responses_manager.remove_unhandled(response)

    async def __handle_open_doc_response(self, websocket, responses_manager,
                                         response):

        doc_handle = response.get('result').get('qReturn').get('qHandle')
        responses_manager.set_handle(doc_handle, self.__DOC_HANDLE)
        follow_up_req_id = await self._send_get_all_infos_request(
            websocket, doc_handle)
        responses_manager.add_pending_id(follow_up_req_id, self._GET_ALL_INFOS)

    async def __handle_get_all_infos_response(self, websocket,
                                              responses_manager, response):

        all_infos = response.get('result').get('qInfos')
        doc_handle = responses_manager.get_handle(self.__DOC_HANDLE)
        master_obj_ids = [
            info.get('qId')
            for info in all_infos
            if 'masterobject' == info.get('qType')
        ]
        follow_up_req_ids = await asyncio.gather(*[
            self.__send_get_object_request(websocket, doc_handle,
                                           master_obj_id)
            for master_obj_id in master_obj_ids
        ])
        responses_manager.add_pending_ids(follow_up_req_ids, self.__GET_OBJECT)

    async def __handle_get_object_response(self, websocket, responses_manager,
                                           response):

        object_handle = response.get('result').get('qReturn').get('qHandle')
        follow_up_req_id = await self.__send_get_properties_request(
            websocket, object_handle)
        responses_manager.add_pending_id(follow_up_req_id,
                                         self.__GET_PROPERTIES)

    async def __send_get_object_request(self, websocket, doc_handle,
                                        object_id):
        """Sends a Get Object Interface request.

        Returns:
            The request id.
        """
        request_id = self._generate_request_id()
        await websocket.send(
            json.dumps({
                'handle': doc_handle,
                'method': self.__GET_OBJECT,
                'params': {
                    'qId': object_id,
                },
                'id': request_id,
            }))

        logging.debug('Get Object Interface request sent: %d', request_id)
        return request_id

    async def __send_get_properties_request(self, websocket, object_handle):
        """Sends a Get Object Properties request.

        Returns:
            The request id.
        """
        request_id = self._generate_request_id()
        await websocket.send(
            json.dumps({
                'handle': object_handle,
                'method': self.__GET_PROPERTIES,
                'params': {},
                'id': request_id,
            }))

        logging.debug('Get Object Properties request sent: %d', request_id)
        return request_id



#!/usr/bin/python
#
# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re
import six
import unicodedata

from google.datacatalog_connectors.qlik.prepare import constants


class DynamicPropertiesHelper:
    __ASCII_CHARACTER_ENCODING = 'ASCII'

    @classmethod
    def make_display_name_for_custom_property_value_tag_template(
            cls, definition_metadata, value):

        display_name = f'Qlik Custom Property' \
                       f' - {definition_metadata.get("name")} - {value}'
        return cls.normalize_tag_template_display_name(display_name)

    @classmethod
    def make_id_for_custom_property_value_tag_template(cls,
                                                       definition_metadata,
                                                       value):

        definition_id = definition_metadata.get("id")
        definition_id_start = definition_id[:definition_id.find('-')]
        generated_id = f'{constants.TAG_TEMPLATE_ID_PREFIX_CUSTOM_PROPERTY}' \
                       f'{definition_id_start}__{value}'
        return cls.normalize_tag_template_id(generated_id)

    @classmethod
    def normalize_tag_template_display_name(cls, source_name):
        formatted_name = re.sub(
            r'[^\w\- ]+', '_',
            cls.__normalize_ascii_chars(source_name).strip())
        return formatted_name

    @classmethod
    def normalize_tag_template_id(cls, source_id):
        lower_case_id = source_id.lower().strip() if source_id else ''
        formatted_id = re.sub(
            r'[^a-z0-9_]+', '_',
            cls.__normalize_ascii_chars(lower_case_id.strip()))
        return formatted_id[:64] if len(formatted_id) > 64 else formatted_id

    @classmethod
    def __normalize_ascii_chars(cls, source_string):
        encoding = cls.__ASCII_CHARACTER_ENCODING
        normalized = unicodedata.normalize(
            'NFKD', source_string
            if isinstance(source_string, six.string_types) else u'')
        encoded = normalized.encode(encoding, 'ignore')
        return encoded.decode()



#!/usr/bin/python
#
# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from google.datacatalog_connectors.commons import prepare

from google.datacatalog_connectors.qlik.prepare import constants


class EntryRelationshipMapper(prepare.BaseEntryRelationshipMapper):
    __APP = constants.USER_SPECIFIED_TYPE_APP
    __CUSTOM_PROPERTY_DEFINITION = \
        constants.USER_SPECIFIED_TYPE_CUSTOM_PROPERTY_DEFINITION
    __DIMENSION = constants.USER_SPECIFIED_TYPE_DIMENSION
    __MEASURE = constants.USER_SPECIFIED_TYPE_MEASURE
    __VISUALIZATION = constants.USER_SPECIFIED_TYPE_VISUALIZATION
    __SHEET = constants.USER_SPECIFIED_TYPE_SHEET
    __STREAM = constants.USER_SPECIFIED_TYPE_STREAM

    def fulfill_tag_fields(self, assembled_entries):
        resolvers = (
            self.__resolve_app_mappings,
            self.__resolve_dimension_mappings,
            self.__resolve_measure_mappings,
            self.__resolve_sheet_mappings,
            self.__resolve_stream_mappings,
            self.__resolve_visualization_mappings,
        )

        self._fulfill_tag_fields(assembled_entries, resolvers)

    @classmethod
    def __resolve_app_mappings(cls, assembled_entries, id_name_pairs):
        app_assembled_entries = [
            assembled_entry for assembled_entry in assembled_entries
            if assembled_entry.entry.user_specified_type == cls.__APP
        ]
        for assembled_entry in app_assembled_entries:
            cls._map_related_entry(assembled_entry,
                                   cls.__CUSTOM_PROPERTY_DEFINITION,
                                   'property_definition_id',
                                   'property_definition_entry', id_name_pairs)
            cls._map_related_entry(assembled_entry, cls.__STREAM, 'stream_id',
                                   'stream_entry', id_name_pairs)

    @classmethod
    def __resolve_dimension_mappings(cls, assembled_entries, id_name_pairs):
        dimension_assembled_entries = [
            assembled_entry for assembled_entry in assembled_entries
            if assembled_entry.entry.user_specified_type == cls.__DIMENSION
        ]
        for assembled_entry in dimension_assembled_entries:
            cls._map_related_entry(assembled_entry, cls.__APP, 'app_id',
                                   'app_entry', id_name_pairs)

    @classmethod
    def __resolve_measure_mappings(cls, assembled_entries, id_name_pairs):
        measure_assembled_entries = [
            assembled_entry for assembled_entry in assembled_entries
            if assembled_entry.entry.user_specified_type == cls.__MEASURE
        ]
        for assembled_entry in measure_assembled_entries:
            cls._map_related_entry(assembled_entry, cls.__APP, 'app_id',
                                   'app_entry', id_name_pairs)

    @classmethod
    def __resolve_sheet_mappings(cls, assembled_entries, id_name_pairs):
        sheet_assembled_entries = [
            assembled_entry for assembled_entry in assembled_entries
            if assembled_entry.entry.user_specified_type == cls.__SHEET
        ]
        for assembled_entry in sheet_assembled_entries:
            cls._map_related_entry(assembled_entry, cls.__APP, 'app_id',
                                   'app_entry', id_name_pairs)

    @classmethod
    def __resolve_stream_mappings(cls, assembled_entries, id_name_pairs):
        stream_assembled_entries = [
            assembled_entry for assembled_entry in assembled_entries
            if assembled_entry.entry.user_specified_type == cls.__STREAM
        ]
        for assembled_entry in stream_assembled_entries:
            cls._map_related_entry(assembled_entry,
                                   cls.__CUSTOM_PROPERTY_DEFINITION,
                                   'property_definition_id',
                                   'property_definition_entry', id_name_pairs)

    @classmethod
    def __resolve_visualization_mappings(cls, assembled_entries,
                                         id_name_pairs):

        viz_assembled_entries = [
            assembled_entry for assembled_entry in assembled_entries
            if assembled_entry.entry.user_specified_type == cls.__VISUALIZATION
        ]
        for assembled_entry in viz_assembled_entries:
            cls._map_related_entry(assembled_entry, cls.__APP, 'app_id',
                                   'app_entry', id_name_pairs)



#!/usr/bin/python
#
# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import abc
import asyncio
import json
import logging
import threading

import jmespath
from urllib.parse import urlparse
import websockets

from google.datacatalog_connectors.qlik.scrape import constants


class BaseEngineAPIHelper(abc.ABC):
    """The base class for all Engine API Helpers."""

    # Methods to be used in the requests.
    _GET_ALL_INFOS = 'GetAllInfos'
    _OPEN_DOC = 'OpenDoc'

    def __init__(self, server_address, auth_cookie):
        # The server address starts with an http/https scheme. The below
        # statement replaces the original scheme with 'wss', which is used for
        # secure websockets communication.
        self.__base_api_endpoint = f'wss://{urlparse(server_address).hostname}'
        self.__auth_cookie = auth_cookie
        self.__common_headers = {
            constants.XRFKEY_HEADER_NAME: constants.XRFKEY,
        }
        self.__requests_counter = 0
        self.__requests_counter_thread_lock = threading.Lock()

    def _connect_websocket(self, app_id):
        """Opens a websocket connection.

        Args:
            app_id:
              An App ID to be appended to the Engine API URLs when making
              websocket connections and then get an isolated Qlik Engine
              session for each Qlik App (see [Connecting to the Qlik Engine
              JSON API](https://help.qlik.com/en-US/sense-developer/September2020/Subsystems/EngineAPI/Content/Sense_EngineAPI/GettingStarted/connecting-to-engine-api.htm)).  # noqa E510

        Returns:
            An awaiting function that yields a :class:`WebSocketClientProtocol`
            which can then be used to send and receive messages.
        """
        uri = f'{self.__base_api_endpoint}/app/{app_id}' \
              f'?Xrfkey={constants.XRFKEY}'

        auth_cookie = self.__auth_cookie
        headers = self.__common_headers.copy()
        # Format the header value as <key>=<value> string.
        headers['Cookie'] = f'{auth_cookie.name}={auth_cookie.value}'

        return websockets.connect(uri=uri, extra_headers=headers)

    def _generate_request_id(self):
        with self.__requests_counter_thread_lock:
            self.__requests_counter += 1
            return self.__requests_counter

    @classmethod
    def _run_until_complete(cls, future):
        event_loop = asyncio.new_event_loop()
        try:
            return event_loop.run_until_complete(future)
        except asyncio.TimeoutError:
            cls.__handle_event_loop_exec_timeout(event_loop)
            raise

    @classmethod
    def __handle_event_loop_exec_timeout(cls, event_loop):
        logging.warning(
            'Timeout reached during the websocket communication session.')
        for task in asyncio.Task.all_tasks(loop=event_loop):
            task.cancel()
        event_loop.stop()

    async def _start_websocket_communication(self, websocket, app_id,
                                             responses_manager):

        request_id = \
            await self.__send_open_doc_request(websocket, app_id)
        responses_manager.add_pending_id(request_id, self._OPEN_DOC)

    @classmethod
    async def _consume_messages(cls, websocket, responses_manager,
                                result_method, result_path):

        results = []
        async for message in websocket:
            response = json.loads(message)
            response_id = response.get('id')
            if not response_id:
                cls.__handle_generic_api_response(response)
                continue

            logging.debug('Response received: %d', response_id)
            if responses_manager.is_pending(response_id, result_method):
                result = jmespath.search(result_path, response)
                if isinstance(result, list):
                    results.extend(result)
                else:
                    results.append(result)
            else:
                responses_manager.add_unhandled(response)

            responses_manager.remove_pending_id(response_id)
            responses_manager.notify_new_response()

        return results

    @classmethod
    def __handle_generic_api_response(cls, response):
        cls.__handle_error_api_response(response)

    @classmethod
    def __handle_error_api_response(cls, response):
        method = response.get('method')
        if 'OnMaxParallelSessionsExceeded' == method:
            message = response.get('params').get('message')
            logging.warning(message)
            raise Exception(message)

    @classmethod
    async def _produce_messages(cls, websocket, responses_manager, producer):
        while not responses_manager.were_all_precessed():
            if not responses_manager.is_there_response_notification():
                await responses_manager.wait_for_responses()
                responses_manager.clear_response_notifications()
            for response in responses_manager.get_all_unhandled():
                await producer(websocket, responses_manager, response)

        # Closes the websocket when there is no further response to process.
        await websocket.close()

    @classmethod
    async def _handle_websocket_communication(cls, consumer_future,
                                              producer_future):

        # The 'results' array is expected to have two elements. The first one
        # stores the result of the consumer, which means the object to be
        # returned on a successfull execution. The second one stores the result
        # of the producer and can be ignored.
        results = await asyncio.gather(*[consumer_future, producer_future])
        return results[0]

    async def __send_open_doc_request(self, websocket, app_id):
        """Sends a Open Doc (aka App) Interface request.

        Returns:
            The request id.
        """
        request_id = self._generate_request_id()
        await websocket.send(
            json.dumps({
                'handle': -1,
                'method': self._OPEN_DOC,
                'params': [app_id],
                'id': request_id
            }))

        logging.debug('Open Doc Interface request sent: %d', request_id)
        return request_id

    async def _send_get_all_infos_request(self, websocket, doc_handle):
        """Sends a Get All Infos request.

        Returns:
            The request id.
        """
        request_id = self._generate_request_id()
        await websocket.send(
            json.dumps({
                'handle': doc_handle,
                'method': self._GET_ALL_INFOS,
                'params': {},
                'id': request_id,
            }))

        logging.debug('Get All Infos request sent: %d', request_id)
        return request_id



#!/usr/bin/python
#
# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import random
import string

JSON_CONTENT_TYPE = 'application/json'

# The Qlik Sense Proxy Service (QPS) manages the Qlik Sense authentication,
# session handling, and load balancing. This constant is used to store the name
# of the cookie issue to a user after authenticating in QPS.
QPS_SESSION_COOKIE_PREFIX = 'X-Qlik-Session'

WINDOWS_USER_AGENT = 'Windows'

# XRFKEY is an arbitrary 16-char-length string composed of letters and digits.
# The below code generates a unique string per connector execution just avoid
# hardcoded values.
XRFKEY = ''.join(random.choices(string.ascii_letters + string.digits, k=16))
XRFKEY_HEADER_NAME = 'x-Qlik-Xrfkey'



#!/usr/bin/python
#
# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
import logging
import sys

from google.datacatalog_connectors.qlik import sync


class Qlik2DataCatalogCli:
    __DEFAULT_DATACATALOG_LOCATION_ID = 'us'
    __DEFAULT_QLIK_DOMAIN = '.'

    @classmethod
    def run(cls, argv):
        cls.__setup_logging()

        args = cls._parse_args(argv)
        args.func(args)

    @classmethod
    def __setup_logging(cls):
        logging.basicConfig(level=logging.INFO)

    @classmethod
    def _parse_args(cls, argv):
        parser = argparse.ArgumentParser(
            description=__doc__,
            formatter_class=argparse.RawDescriptionHelpFormatter)

        parser.add_argument('--qlik-server', help='Qlik server', required=True)
        parser.add_argument('--qlik-ad-domain',
                            help='Qlik Active Directory domain',
                            default=cls.__DEFAULT_QLIK_DOMAIN)
        parser.add_argument('--qlik-username',
                            help='Qlik username',
                            required=True)
        parser.add_argument('--qlik-password',
                            help='Qlik password',
                            required=True)
        parser.add_argument('--datacatalog-project-id',
                            help='Google Cloud Project ID',
                            required=True)
        parser.add_argument('--datacatalog-location-id',
                            help='Location ID to be used Google Data Catalog'
                            'to store the metadata',
                            default=cls.__DEFAULT_DATACATALOG_LOCATION_ID)

        parser.set_defaults(func=cls.__run_synchronizer)

        return parser.parse_args(argv)

    @classmethod
    def __run_synchronizer(cls, args):
        sync.MetadataSynchronizer(
            qlik_server_address=args.qlik_server,
            qlik_ad_domain=args.qlik_ad_domain,
            qlik_username=args.qlik_username,
            qlik_password=args.qlik_password,
            datacatalog_project_id=args.datacatalog_project_id,
            datacatalog_location_id=args.datacatalog_location_id).run()


def main():
    argv = sys.argv
    Qlik2DataCatalogCli.run(argv[1:] if len(argv) > 0 else argv)



#!/usr/bin/python
#
# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging

from google.datacatalog_connectors.qlik.scrape import \
    engine_api_scraper, repository_services_api_helper


class MetadataScraper:
    """A Facade that provides a simplified interface for the Qlik services,
    comprising the interactions between Qlik Sense Proxy Service (QPS), Qlik
    Sense Repository Service (QRS), and Qlik Engine JSON API.
    """

    def __init__(self, server_address, ad_domain, username, password):
        self.__qrs_api_helper = \
            repository_services_api_helper.RepositoryServicesAPIHelper(
                server_address, ad_domain, username, password)
        self.__engine_api_scraper = engine_api_scraper.EngineAPIScraper(
            server_address, ad_domain, username, password)

    def scrape_all_apps(self):
        self.__log_scrape_start('Scraping all Apps...')
        apps = self.__qrs_api_helper.get_full_app_list()

        logging.info('  %s Apps found:', len(apps))
        for app in apps:
            logging.info(
                '    - %s :: %s [%s]',
                app.get('stream').get('name')
                if app.get('published') else 'NOT PUBLISHED!', app.get('name'),
                app.get('id'))

        return apps

    def scrape_all_custom_property_definitions(self):
        self.__log_scrape_start('Scraping all Custom Property Definitions...')
        defs = self.__qrs_api_helper.get_full_custom_property_definition_list()

        logging.info('  %s Custom Property Definitions found:', len(defs))
        for defintion in defs:
            logging.info('    - %s [%s]', defintion.get('name'),
                         defintion.get('id'))

        return defs

    def scrape_all_streams(self):
        self.__log_scrape_start('Scraping all Streams...')
        streams = self.__qrs_api_helper.get_full_stream_list()

        logging.info('  %s Streams found:', len(streams))
        for stream in streams:
            logging.info('    - %s [%s]', stream.get('name'), stream.get('id'))

        return streams

    def scrape_dimensions(self, app_metadata):
        self.__log_scrape_start(
            'Scraping Dimensions (Master Items) from the'
            ' "%s" App...', app_metadata.get('name'))
        dimensions = self.__engine_api_scraper.get_dimensions(
            app_metadata.get('id')) or []

        logging.info('  %s Dimensions found:', len(dimensions))
        for dimension in dimensions:
            q_meta_def = dimension.get('qMetaDef')
            logging.info('    - %s [%s]', q_meta_def.get('title'),
                         dimension.get('qInfo').get('qId'))

        return dimensions

    def scrape_measures(self, app_metadata):
        self.__log_scrape_start(
            'Scraping Measures (Master Items) from the'
            ' "%s" App...', app_metadata.get('name'))
        measures = self.__engine_api_scraper.get_measures(
            app_metadata.get('id')) or []

        logging.info('  %s Measures found:', len(measures))
        for measure in measures:
            q_meta_def = measure.get('qMetaDef')
            logging.info('    - %s [%s]', q_meta_def.get('title'),
                         measure.get('qInfo').get('qId'))

        return measures

    def scrape_sheets(self, app_metadata):
        self.__log_scrape_start('Scraping Sheets from the "%s" App...',
                                app_metadata.get('name'))
        sheets = self.__engine_api_scraper.get_sheets(
            app_metadata.get('id')) or []

        logging.info('  %s Sheets found:', len(sheets))
        for sheet in sheets:
            q_meta = sheet.get('qMeta')
            logging.info('    - %s%s [%s]',
                         '' if q_meta.get('published') else 'NOT PUBLISHED! ',
                         q_meta.get('title'),
                         sheet.get('qInfo').get('qId'))

        return sheets

    def scrape_visualizations(self, app_metadata):
        self.__log_scrape_start(
            'Scraping Visualizations (Master Items) from the'
            ' "%s" App...', app_metadata.get('name'))
        visualizations = self.__engine_api_scraper.get_visualizations(
            app_metadata.get('id')) or []

        logging.info('  %s Visualizations found:', len(visualizations))
        for visualization in visualizations:
            q_meta_def = visualization.get('qMetaDef')
            logging.info('    - %s [%s]', q_meta_def.get('title'),
                         visualization.get('qInfo').get('qId'))

        return visualizations

    @classmethod
    def __log_scrape_start(cls, message, *args):
        logging.info('')
        logging.info(message, *args)
        logging.info('-------------------------------------------------')



#!/usr/bin/python
#
# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from google.datacatalog_connectors.commons import prepare

from google.datacatalog_connectors.qlik.prepare import \
    constants, datacatalog_entry_factory, datacatalog_tag_factory


class AssembledEntryFactory:
    __datacatalog_entry_factory: datacatalog_entry_factory\
        .DataCatalogEntryFactory

    def __init__(self, project_id, location_id, entry_group_id,
                 user_specified_system, site_url):

        self.__datacatalog_entry_factory = datacatalog_entry_factory \
            .DataCatalogEntryFactory(
                project_id, location_id, entry_group_id, user_specified_system,
                site_url)

        self.__datacatalog_tag_factory = \
            datacatalog_tag_factory.DataCatalogTagFactory(site_url)

    def make_assembled_entry_for_custom_property_def(
            self, custom_property_def_metadata, tag_template):

        entry_id, entry = self.__datacatalog_entry_factory\
            .make_entry_for_custom_property_definition(
                custom_property_def_metadata)

        tags = []
        if tag_template:
            tags.append(
                self.__datacatalog_tag_factory.
                make_tag_for_custom_property_definition(
                    tag_template, custom_property_def_metadata))

        return prepare.AssembledEntryData(entry_id, entry, tags)

    def make_assembled_entries_for_stream(self, stream_metadata,
                                          tag_templates_dict):

        assembled_entries = [
            self.__make_assembled_entry_for_stream(stream_metadata,
                                                   tag_templates_dict)
        ]

        assembled_entries.extend(
            self.__make_assembled_entries_for_apps(stream_metadata.get('apps'),
                                                   tag_templates_dict))

        return assembled_entries

    def __make_assembled_entry_for_stream(self, stream_metadata,
                                          tag_templates_dict):

        entry_id, entry = \
            self.__datacatalog_entry_factory.make_entry_for_stream(
                stream_metadata)

        stream_tag_template = tag_templates_dict.get(
            constants.TAG_TEMPLATE_ID_STREAM)

        tags = []
        if stream_tag_template:
            tags.append(
                self.__datacatalog_tag_factory.make_tag_for_stream(
                    stream_tag_template, stream_metadata))

        tags.extend(
            self.__datacatalog_tag_factory.make_tags_for_custom_properties(
                tag_templates_dict, stream_metadata.get('customProperties')))

        return prepare.AssembledEntryData(entry_id, entry, tags)

    def __make_assembled_entries_for_apps(self, apps_metadata,
                                          tag_templates_dict):

        assembled_entries = []

        if not apps_metadata:
            return assembled_entries

        for app_metadata in apps_metadata:
            assembled_entries.append(
                self.__make_assembled_entry_for_app(app_metadata,
                                                    tag_templates_dict))

            assembled_entries.extend(
                self.__make_assembled_entries_for_dimensions(
                    app_metadata.get('dimensions'), tag_templates_dict))

            assembled_entries.extend(
                self.__make_assembled_entries_for_measures(
                    app_metadata.get('measures'), tag_templates_dict))

            assembled_entries.extend(
                self.__make_assembled_entries_for_visualizations(
                    app_metadata.get('visualizations'), tag_templates_dict))

            assembled_entries.extend(
                self.__make_assembled_entries_for_sheets(
                    app_metadata.get('sheets'), tag_templates_dict))

        return assembled_entries

    def __make_assembled_entry_for_app(self, app_metadata, tag_templates_dict):
        entry_id, entry = \
            self.__datacatalog_entry_factory.make_entry_for_app(app_metadata)

        app_tag_template = tag_templates_dict.get(
            constants.TAG_TEMPLATE_ID_APP)

        tags = []
        if app_tag_template:
            tags.append(
                self.__datacatalog_tag_factory.make_tag_for_app(
                    app_tag_template, app_metadata))

        tags.extend(
            self.__datacatalog_tag_factory.make_tags_for_custom_properties(
                tag_templates_dict, app_metadata.get('customProperties')))

        return prepare.AssembledEntryData(entry_id, entry, tags)

    def __make_assembled_entries_for_dimensions(self, dimensions_metadata,
                                                tag_templates_dict):

        return [
            self.__make_assembled_entry_for_dimension(dimension_metadata,
                                                      tag_templates_dict)
            for dimension_metadata in dimensions_metadata
        ] if dimensions_metadata else []

    def __make_assembled_entry_for_dimension(self, dimension_metadata,
                                             tag_templates_dict):

        entry_id, entry = \
            self.__datacatalog_entry_factory.make_entry_for_dimension(
                dimension_metadata)

        tag_template = tag_templates_dict.get(
            constants.TAG_TEMPLATE_ID_DIMENSION)

        tags = []
        if tag_template:
            tags.append(
                self.__datacatalog_tag_factory.make_tag_for_dimension(
                    tag_template, dimension_metadata))

        return prepare.AssembledEntryData(entry_id, entry, tags)

    def __make_assembled_entries_for_measures(self, measures_metadata,
                                              tag_templates_dict):

        return [
            self.__make_assembled_entry_for_measure(measure_metadata,
                                                    tag_templates_dict)
            for measure_metadata in measures_metadata
        ] if measures_metadata else []

    def __make_assembled_entry_for_measure(self, measure_metadata,
                                           tag_templates_dict):

        entry_id, entry = \
            self.__datacatalog_entry_factory.make_entry_for_measure(
                measure_metadata)

        tag_template = tag_templates_dict.get(
            constants.TAG_TEMPLATE_ID_MEASURE)

        tags = []
        if tag_template:
            tags.append(
                self.__datacatalog_tag_factory.make_tag_for_measure(
                    tag_template, measure_metadata))

        return prepare.AssembledEntryData(entry_id, entry, tags)

    def __make_assembled_entries_for_sheets(self, sheets_metadata,
                                            tag_templates_dict):

        return [
            self.__make_assembled_entry_for_sheet(sheet_metadata,
                                                  tag_templates_dict)
            for sheet_metadata in sheets_metadata
        ] if sheets_metadata else []

    def __make_assembled_entry_for_sheet(self, sheet_metadata,
                                         tag_templates_dict):

        entry_id, entry = \
            self.__datacatalog_entry_factory.make_entry_for_sheet(
                sheet_metadata)

        tag_template = tag_templates_dict.get(constants.TAG_TEMPLATE_ID_SHEET)

        tags = []
        if tag_template:
            tags.append(
                self.__datacatalog_tag_factory.make_tag_for_sheet(
                    tag_template, sheet_metadata))

        return prepare.AssembledEntryData(entry_id, entry, tags)

    def __make_assembled_entries_for_visualizations(self,
                                                    visualizations_metadata,
                                                    tag_templates_dict):

        return [
            self.__make_assembled_entry_for_visualization(
                visualization_metadata, tag_templates_dict)
            for visualization_metadata in visualizations_metadata
        ] if visualizations_metadata else []

    def __make_assembled_entry_for_visualization(self, visualization_metadata,
                                                 tag_templates_dict):

        entry_id, entry = \
            self.__datacatalog_entry_factory.make_entry_for_visualization(
                visualization_metadata)

        tag_template = tag_templates_dict.get(
            constants.TAG_TEMPLATE_ID_VISUALIZATION)

        tags = []
        if tag_template:
            tags.append(
                self.__datacatalog_tag_factory.make_tag_for_visualization(
                    tag_template, visualization_metadata))

        return prepare.AssembledEntryData(entry_id, entry, tags)



#!/usr/bin/python
#
# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging

import requests
from requests_ntlm import HttpNtlmAuth

from google.datacatalog_connectors.qlik.scrape import constants


class Authenticator:

    @classmethod
    def get_qps_session_cookie_windows_auth(cls, ad_domain, username, password,
                                            auth_url):

        # Set up user credentials
        user_auth = HttpNtlmAuth(username=f'{ad_domain}\\{username}',
                                 password=password)

        headers = {constants.XRFKEY_HEADER_NAME: constants.XRFKEY}

        response = requests.get(url=auth_url, auth=user_auth, headers=headers)

        if not response.cookies:
            cls.__log_authentication_failure()
            return

        try:
            return next(
                cookie for cookie in response.cookies
                if cookie.name.startswith(constants.QPS_SESSION_COOKIE_PREFIX))
        except StopIteration:
            cls.__log_authentication_failure()

    @classmethod
    def __log_authentication_failure(cls):
        logging.warning(
            'Authentication failed! %s cookie not found in the response.',
            constants.QPS_SESSION_COOKIE_PREFIX)



#!/usr/bin/python
#
# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from datetime import datetime

from google.cloud import datacatalog
from google.datacatalog_connectors.commons import prepare

from google.datacatalog_connectors.qlik.prepare import \
    constants, dynamic_properties_helper as dph


class DataCatalogTagFactory(prepare.BaseTagFactory):
    __INCOMING_TIMESTAMP_UTC_FORMAT = '%Y-%m-%dT%H:%M:%S.%fZ'
    __QLIK_TO_DC_DIM_GROUPING_MAPPING = {
        constants.DIMENSION_GROUPING_SINGLE_QLIK:
            constants.DIMENSION_GROUPING_SINGLE_TAG_FIELD,
        constants.DIMENSION_GROUPING_DRILL_DOWN_QLIK:
            constants.DIMENSION_GROUPING_DRILL_DOWN_TAG_FIELD,
    }

    def __init__(self, site_url):
        self.__site_url = site_url

    def make_tag_for_app(self, tag_template, app_metadata):
        tag = datacatalog.Tag()

        tag.template = tag_template.name

        self._set_string_field(tag, 'id', app_metadata.get('id'))

        owner = app_metadata.get('owner')
        if owner:
            owner_user_dir = owner.get('userDirectory')
            owner_user_id = owner.get('userId')
            if owner_user_dir and owner_user_id:
                self._set_string_field(tag, 'owner_username',
                                       f'{owner_user_dir}\\\\{owner_user_id}')

            self._set_string_field(tag, 'owner_name', owner.get('name'))

        self._set_string_field(tag, 'modified_by_username',
                               app_metadata.get('modifiedByUserName'))

        published = app_metadata.get('published')
        self._set_bool_field(tag, 'published', published)

        publish_time = app_metadata.get('publishTime')
        # The publish time may come with a value such as
        # '1753-01-01T00:00:00.000Z' when the app has not been published,
        # so both inputs are tested before setting the tag field.
        if published and publish_time:
            self._set_timestamp_field(
                tag, 'publish_time',
                datetime.strptime(publish_time,
                                  self.__INCOMING_TIMESTAMP_UTC_FORMAT))

        last_reload_time = app_metadata.get('lastReloadTime')
        if last_reload_time:
            self._set_timestamp_field(
                tag, 'last_reload_time',
                datetime.strptime(last_reload_time,
                                  self.__INCOMING_TIMESTAMP_UTC_FORMAT))

        stream_metadata = app_metadata.get('stream')
        if stream_metadata:
            self._set_string_field(tag, 'stream_id', stream_metadata.get('id'))
            self._set_string_field(tag, 'stream_name',
                                   stream_metadata.get('name'))

        file_size = app_metadata.get('fileSize')
        if file_size is not None:
            self._set_string_field(
                tag, 'file_size',
                self.__get_human_readable_size_value(file_size))

        if app_metadata.get('thumbnail'):
            self._set_string_field(
                tag, 'thumbnail',
                f'{self.__site_url}{app_metadata.get("thumbnail")}')

        self._set_string_field(tag, 'saved_in_product_version',
                               app_metadata.get('savedInProductVersion'))

        self._set_string_field(tag, 'migration_hash',
                               app_metadata.get('migrationHash'))

        self._set_double_field(tag, 'availability_status',
                               app_metadata.get('availabilityStatus'))

        self._set_string_field(tag, 'site_url', self.__site_url)

        return tag

    def make_tags_for_custom_properties(self, tag_templates_dict,
                                        custom_properties):

        tags = []

        if not (tag_templates_dict and custom_properties):
            return tags

        for property_metadata in custom_properties:
            definition = property_metadata.get('definition')
            value = property_metadata.get('value')
            template_id = dph.DynamicPropertiesHelper\
                .make_id_for_custom_property_value_tag_template(
                    definition, value)
            property_value_tag_template = tag_templates_dict.get(template_id)

            if not property_value_tag_template:
                continue

            tags.append(
                self.make_tag_for_custom_property(property_value_tag_template,
                                                  property_metadata))

        return tags

    def make_tag_for_custom_property(self, tag_template,
                                     custom_property_metadata):

        tag = datacatalog.Tag()

        tag.template = tag_template.name

        self._set_string_field(tag, 'id', custom_property_metadata.get('id'))

        created_date = custom_property_metadata.get('createdDate')
        if created_date:
            self._set_timestamp_field(
                tag, 'created_date',
                datetime.strptime(created_date,
                                  self.__INCOMING_TIMESTAMP_UTC_FORMAT))

        modified_date = custom_property_metadata.get('modifiedDate')
        if modified_date:
            self._set_timestamp_field(
                tag, 'modified_date',
                datetime.strptime(modified_date,
                                  self.__INCOMING_TIMESTAMP_UTC_FORMAT))

        self._set_string_field(
            tag, 'modified_by_username',
            custom_property_metadata.get('modifiedByUserName'))

        self._set_string_field(tag, 'value',
                               custom_property_metadata.get('value'))

        definition = custom_property_metadata.get('definition')
        if definition:
            self._set_string_field(tag, 'property_definition_id',
                                   definition.get('id'))
            self._set_string_field(tag, 'property_name',
                                   definition.get('name'))

        self._set_string_field(tag, 'site_url', self.__site_url)

        return tag

    def make_tag_for_custom_property_definition(self, tag_template,
                                                custom_property_def_metadata):

        tag = datacatalog.Tag()

        tag.template = tag_template.name

        self._set_string_field(tag, 'id',
                               custom_property_def_metadata.get('id'))

        self._set_string_field(
            tag, 'modified_by_username',
            custom_property_def_metadata.get('modifiedByUserName'))

        self._set_string_field(tag, 'value_type',
                               custom_property_def_metadata.get('valueType'))

        choice_values = custom_property_def_metadata.get('choiceValues') or []
        self._set_string_field(tag, 'choice_values', ', '.join(choice_values))

        object_types = custom_property_def_metadata.get('objectTypes') or []
        self._set_string_field(tag, 'object_types', ', '.join(object_types))

        self._set_string_field(tag, 'site_url', self.__site_url)

        return tag

    def make_tag_for_dimension(self, tag_template, dimension_metadata):
        tag = datacatalog.Tag()

        tag.template = tag_template.name

        self._set_string_field(tag, 'id',
                               dimension_metadata.get('qInfo').get('qId'))

        q_dim = dimension_metadata.get('qDim')
        grouping = self.__QLIK_TO_DC_DIM_GROUPING_MAPPING.get(
            q_dim.get('qGrouping'))
        self.__set_enum_field(tag, 'grouping', grouping)

        self._set_string_field(tag, 'fields',
                               ', '.join(q_dim.get('qFieldDefs')))
        self._set_string_field(tag, 'field_labels',
                               ', '.join(q_dim.get('qFieldLabels')))

        q_meta_def = dimension_metadata.get('qMetaDef')
        self._set_string_field(tag, 'tags', ', '.join(q_meta_def.get('tags')))

        app_metadata = dimension_metadata.get('app')
        if app_metadata:
            self._set_string_field(tag, 'app_id', app_metadata.get('id'))
            self._set_string_field(tag, 'app_name', app_metadata.get('name'))

        self._set_string_field(tag, 'site_url', self.__site_url)

        return tag

    def make_tag_for_measure(self, tag_template, measure_metadata):
        tag = datacatalog.Tag()

        tag.template = tag_template.name

        self._set_string_field(tag, 'id',
                               measure_metadata.get('qInfo').get('qId'))

        q_measure = measure_metadata.get('qMeasure')

        self._set_string_field(tag, 'expression', q_measure.get('qDef'))
        self._set_string_field(tag, 'label_expression',
                               q_measure.get('qLabelExpression'))
        self._set_bool_field(tag, 'is_custom_formatted',
                             q_measure.get('isCustomFormatted'))

        q_meta_def = measure_metadata.get('qMetaDef')
        self._set_string_field(tag, 'tags', ', '.join(q_meta_def.get('tags')))

        app_metadata = measure_metadata.get('app')
        if app_metadata:
            self._set_string_field(tag, 'app_id', app_metadata.get('id'))
            self._set_string_field(tag, 'app_name', app_metadata.get('name'))

        self._set_string_field(tag, 'site_url', self.__site_url)

        return tag

    def make_tag_for_sheet(self, tag_template, sheet_metadata):
        tag = datacatalog.Tag()

        tag.template = tag_template.name

        self._set_string_field(tag, 'id',
                               sheet_metadata.get('qInfo').get('qId'))

        q_meta = sheet_metadata.get('qMeta')
        owner = q_meta.get('owner')
        if owner:
            owner_user_dir = owner.get('userDirectory')
            owner_user_id = owner.get('userId')
            if owner_user_dir and owner_user_id:
                self._set_string_field(tag, 'owner_username',
                                       f'{owner_user_dir}\\\\{owner_user_id}')

            self._set_string_field(tag, 'owner_name', owner.get('name'))

        published = q_meta.get('published')
        self._set_bool_field(tag, 'published', published)

        publish_time = q_meta.get('publishTime')
        # The publish time may come with a value such as
        # '1753-01-01T00:00:00.000Z' when the sheet has not been published,
        # so both inputs are tested before setting the tag field.
        if published and publish_time:
            self._set_timestamp_field(
                tag, 'publish_time',
                datetime.strptime(publish_time,
                                  self.__INCOMING_TIMESTAMP_UTC_FORMAT))

        self._set_bool_field(tag, 'approved', q_meta.get('approved'))

        app_metadata = sheet_metadata.get('app')
        if app_metadata:
            self._set_string_field(tag, 'app_id', app_metadata.get('id'))
            self._set_string_field(tag, 'app_name', app_metadata.get('name'))

        self._set_string_field(tag, 'source_object',
                               q_meta.get('sourceObject'))

        self._set_string_field(tag, 'draft_object', q_meta.get('draftObject'))

        self._set_string_field(tag, 'site_url', self.__site_url)

        return tag

    def make_tag_for_stream(self, tag_template, stream_metadata):
        tag = datacatalog.Tag()

        tag.template = tag_template.name

        self._set_string_field(tag, 'id', stream_metadata.get('id'))

        owner = stream_metadata.get('owner')
        if owner:
            owner_user_dir = owner.get('userDirectory')
            owner_user_id = owner.get('userId')
            if owner_user_dir and owner_user_id:
                self._set_string_field(tag, 'owner_username',
                                       f'{owner_user_dir}\\\\{owner_user_id}')

            self._set_string_field(tag, 'owner_name', owner.get('name'))

        self._set_string_field(tag, 'modified_by_username',
                               stream_metadata.get('modifiedByUserName'))

        self._set_string_field(tag, 'site_url', self.__site_url)

        return tag

    def make_tag_for_visualization(self, tag_template, visualization_metadata):
        tag = datacatalog.Tag()

        tag.template = tag_template.name

        self._set_string_field(tag, 'id',
                               visualization_metadata.get('qInfo').get('qId'))

        self._set_string_field(tag, 'title',
                               visualization_metadata.get('title'))
        self._set_string_field(tag, 'subtitle',
                               visualization_metadata.get('subtitle'))
        self._set_string_field(tag, 'footnote',
                               visualization_metadata.get('footnote'))
        self._set_string_field(tag, 'type',
                               visualization_metadata.get('visualization'))

        q_meta_def = visualization_metadata.get('qMetaDef')
        self._set_string_field(tag, 'tags', ', '.join(q_meta_def.get('tags')))

        app_metadata = visualization_metadata.get('app')
        if app_metadata:
            self._set_string_field(tag, 'app_id', app_metadata.get('id'))
            self._set_string_field(tag, 'app_name', app_metadata.get('name'))

        self._set_string_field(tag, 'site_url', self.__site_url)

        return tag

    @classmethod
    def __get_human_readable_size_value(cls, size_bytes):
        """

        :param size_bytes: int or string, in bytes
        :return: human-readable size
        """
        size_val = int(size_bytes)
        units = ['bytes', 'KB', 'MB', 'GB']
        for unit in units:
            if size_val < 1024.0:
                human_readable_space = f'{round(size_val, 2)} {unit}'
                return human_readable_space
            size_val = size_val / 1024.0
        return f'{round(size_val, 2)} TB'

    @classmethod
    def __set_enum_field(cls, tag, field_id, value):
        if value is not None:
            enum_field = datacatalog.TagField()
            enum_field.enum_value.display_name = value
            tag.fields[field_id] = enum_field



#!/usr/bin/python
#
# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio


class WebsocketResponsesManager:
    """Utility class with common features for responses handling over websocket
    communication sessions.

    Attributes:
        __pending_ids:
            A ``list`` containing the ids of the pending responses, which means
            the requests identified by them were sent but not answered yet.
        __api_calls_history:
            A ``dict`` containing a full history of the API calls known by a
            given ``WebsocketResponsesManager`` instance, represented as
            ``reponse-id: method`` items. It is automatically fulfilled when
            pending ids are added to ``__pending_ids``.
        __unhandled_responses:
            A ``list`` containing all response objects that were received but
            not handled yet.
        __new_response_event:
            A signal used by the consumer to notify the producer on the arrival
            of new responses, so the producer can take actions such as sending
            follow up requests.
        __interface_handles:
            A ``dict`` containing the keys and values of the interface handles
            required by a given communication session.
    """

    def __init__(self):
        self.__pending_ids = []
        self.__api_calls_history = {}
        self.__unhandled_responses = []
        self.__new_response_event = asyncio.Event()
        self.__interface_handles = {}

    def add_pending_id(self, response_id, method):
        self.__pending_ids.append(response_id)
        self.__api_calls_history[response_id] = method

    def add_pending_ids(self, response_ids, method):
        for response_id in response_ids:
            self.add_pending_id(response_id, method)

    def remove_pending_id(self, response_id):
        self.__pending_ids.remove(response_id)

    def is_pending(self, response_id, method):
        return response_id in self.__pending_ids and self.is_method(
            response_id, method)

    def is_method(self, response_id, method):
        return method == self.__api_calls_history.get(response_id)

    def add_unhandled(self, response):
        self.__unhandled_responses.append(response)

    def remove_unhandled(self, response):
        self.__unhandled_responses.remove(response)

    def get_all_unhandled(self):
        return self.__unhandled_responses

    def were_all_precessed(self):
        return not self.__pending_ids and not self.__unhandled_responses

    def notify_new_response(self):
        self.__new_response_event.set()

    async def wait_for_responses(self):
        return await self.__new_response_event.wait()

    def is_there_response_notification(self):
        return self.__new_response_event.is_set()

    def clear_response_notifications(self):
        self.__new_response_event.clear()

    def set_handle(self, value, handle_key):
        self.__interface_handles[handle_key] = value

    def get_handle(self, handle_key):
        return self.__interface_handles[handle_key]



#!/usr/bin/python
#
# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import requests
from requests import sessions

from google.datacatalog_connectors.qlik.scrape import authenticator, constants


class RepositoryServicesAPIHelper:
    """Wraps requests to the Qlik Sense Repository Service (QRS).

    The Qlik Sense Repository Service (QRS) manages persistence and
    synchronization of Qlik Sense apps, licensing, security, and service
    configuration data.

    QRS returns full objects when /full is included in the request path;
    otherwise, it returns condensed objects. Full objects are preferred
    because they allow the connector to read a richer metadata set (see [Full
    versus condensed objects](https://help.qlik.com/en-US/sense-developer/September2020/Subsystems/RepositoryServiceAPI/Content/Sense_RepositoryServiceAPI/RepositoryServiceAPI-Connect-API-Full-vs-Condensed-Objects.htm)  # noqa E501
    for more information).

    Attributes:
        __http_session: An HTTP session for the QRS RESP API calls.

    """

    def __init__(self, server_address, ad_domain, username, password):
        self.__server_address = server_address
        self.__ad_domain = ad_domain
        self.__username = username
        self.__password = password

        self.__base_api_endpoint = f'{server_address}/qrs'
        self.__common_headers = {
            'Content-Type': constants.JSON_CONTENT_TYPE,
            'Accept': constants.JSON_CONTENT_TYPE,
            constants.XRFKEY_HEADER_NAME: constants.XRFKEY,
        }

        self.__http_session = None

    def get_full_app_list(self):
        """Get the list of all Apps that can be opened by the current user,
        via the current proxy.

        Returns:
            A list of full App metadata objects.
        """
        return self.__execute_api_call('app/hublist/full')

    def get_full_custom_property_definition_list(self):
        """Get the list of Custom Property Definitions with full metadata from
        a given server.

        Returns:
            A list of full Custom Property Definition metadata objects.
        """
        return self.__execute_api_call('custompropertydefinition/full')

    def get_full_stream_list(self):
        """Get the list of Streams with full metadata from a given server.

        Returns:
            A list of full Stream metadata objects.
        """
        return self.__execute_api_call('stream/full')

    def __execute_api_call(self, resource_path):
        self.__set_up_http_session()

        url = f'{self.__base_api_endpoint}/{resource_path}' \
              f'?Xrfkey={constants.XRFKEY}'

        return self.__http_session.get(url=url,
                                       headers=self.__common_headers).json()

    def __set_up_http_session(self):
        if self.__http_session:
            return

        self.__http_session = sessions.Session()

        windows_auth_url = self.__get_windows_authentication_url(
            self.__http_session)
        qps_session_cookie = authenticator.Authenticator \
            .get_qps_session_cookie_windows_auth(
                ad_domain=self.__ad_domain,
                username=self.__username,
                password=self.__password,
                auth_url=windows_auth_url)

        self.__http_session.cookies.set_cookie(qps_session_cookie)
        logging.debug('QPS session cookie issued for the QRS API: %s',
                      qps_session_cookie)

    def __get_windows_authentication_url(self, session):
        """Get a Windows Authentication url.

        This method sends an unauthenticated request to a well known endpoint
        of the Qlik Sense Repository Service API. The expected response has a
        302 status code and a `Location` header, which is the Windows
        Authentication url.

        Returns:
            A string.
        """
        url = f'{self.__base_api_endpoint}/about?Xrfkey={constants.XRFKEY}'

        # Sets the User-Agent to Windows temporarily to get a Windows
        # Authentication URL that is required by the NTLM authentication flow.
        headers = self.__common_headers.copy()
        headers['User-Agent'] = constants.WINDOWS_USER_AGENT

        response = requests.get(url=url,
                                headers=headers,
                                allow_redirects=False)

        return session.get_redirect_target(response)



#!/usr/bin/python
#
# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# The max length of a Data Catalog Entry ID.
ENTRY_ID_MAX_LENGTH = 64
# The below asset type specific strings are appended to the standard
# ENTRY_ID_PREFIX when generating Qlik Entry IDs.
ENTRY_ID_PART_APP = 'app_'
ENTRY_ID_PART_CUSTOM_PROPERTY_DEFINITION = 'cpd_'
ENTRY_ID_PART_DIMENSION = 'dim_'
ENTRY_ID_PART_MEASURE = 'msr_'
ENTRY_ID_PART_SHEET = 'sht_'
ENTRY_ID_PART_STREAM = 'str_'
ENTRY_ID_PART_VISUALIZATION = 'viz_'
# This is the common prefix for all Qlik Entries.
ENTRY_ID_PREFIX = 'qlik_'
# This constant represents the max length a generated Entry ID can have before
# appending the standard ENTRY_ID_PREFIX to it. The necessary right-most chars
# will be removed from a generated Entry ID if it has more chars than the value
# of this constant.
NO_PREFIX_ENTRY_ID_MAX_LENGTH = ENTRY_ID_MAX_LENGTH - len(ENTRY_ID_PREFIX)

# The ID of the Tag Template created to store additional metadata for the
# App-related Entries.
TAG_TEMPLATE_ID_APP = 'qlik_app_metadata'
# The ID of the Tag Template created to store additional metadata for the
# Custom Property Definition related Entries.
TAG_TEMPLATE_ID_CUSTOM_PROPERTY_DEFINITION = \
    'qlik_custom_property_definition_metadata'
# Prefix for IDs of the Tag Templates created to tag Entries with their
# Custom Properties.
TAG_TEMPLATE_ID_PREFIX_CUSTOM_PROPERTY = 'qlik_cp__'
# The ID of the Tag Template created to store additional metadata for the
# Dimension-related Entries.
TAG_TEMPLATE_ID_DIMENSION = 'qlik_dimension_metadata'
# The ID of the Tag Template created to store additional metadata for the
# Measure-related Entries.
TAG_TEMPLATE_ID_MEASURE = 'qlik_measure_metadata'
# The ID of the Tag Template created to store additional metadata for the
# Sheet-related Entries.
TAG_TEMPLATE_ID_SHEET = 'qlik_sheet_metadata'
# The ID of the Tag Template created to store additional metadata for the
# Stream-related Entries.
TAG_TEMPLATE_ID_STREAM = 'qlik_stream_metadata'
# The ID of the Tag Template created to store additional metadata for the
# Visualization-related Entries.
TAG_TEMPLATE_ID_VISUALIZATION = 'qlik_visualization_metadata'

# The user specified type of the App-related Entries.
USER_SPECIFIED_TYPE_APP = 'app'
# The user specified type of the Custom Property Definition related Entries.
USER_SPECIFIED_TYPE_CUSTOM_PROPERTY_DEFINITION = 'custom_property_definition'
# The user specified type of the Dimension-related Entries.
USER_SPECIFIED_TYPE_DIMENSION = 'dimension'
# The user specified type of the Measure-related Entries.
USER_SPECIFIED_TYPE_MEASURE = 'measure'
# The user specified type of the Sheet-related Entries.
USER_SPECIFIED_TYPE_SHEET = 'sheet'
# The user specified type of the Stream-related Entries.
USER_SPECIFIED_TYPE_STREAM = 'stream'
# The user specified type of the Visualization-related Entries.
USER_SPECIFIED_TYPE_VISUALIZATION = 'visualization'

# The value of the Tag Field that represents a 'Drill down' Dimension.
DIMENSION_GROUPING_DRILL_DOWN_TAG_FIELD = 'Drill down'
# The value of the incoming metadata that represents a 'Drill down' Dimension.
DIMENSION_GROUPING_DRILL_DOWN_QLIK = 'H'
# The value of the Tag Field that represents a 'Single' Dimension.
DIMENSION_GROUPING_SINGLE_TAG_FIELD = 'Single'
# The value of the incoming metadata that represents a 'Single' Dimension.
DIMENSION_GROUPING_SINGLE_QLIK = 'N'



#!/usr/bin/python
#
# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from datetime import datetime

from google.cloud import datacatalog
from google.protobuf import timestamp_pb2
from google.datacatalog_connectors.commons import prepare

from google.datacatalog_connectors.qlik.prepare import constants


class DataCatalogEntryFactory(prepare.BaseEntryFactory):
    __INCOMING_TIMESTAMP_UTC_FORMAT = '%Y-%m-%dT%H:%M:%S.%fZ'

    def __init__(self, project_id, location_id, entry_group_id,
                 user_specified_system, site_url):

        self.__project_id = project_id
        self.__location_id = location_id
        self.__entry_group_id = entry_group_id
        self.__user_specified_system = user_specified_system
        self.__site_url = site_url

    def make_entry_for_app(self, app_metadata):
        entry = datacatalog.Entry()

        generated_id = self.__format_id(constants.ENTRY_ID_PART_APP,
                                        app_metadata.get('id'))
        entry.name = datacatalog.DataCatalogClient.entry_path(
            self.__project_id, self.__location_id, self.__entry_group_id,
            generated_id)

        entry.user_specified_system = self.__user_specified_system
        entry.user_specified_type = constants.USER_SPECIFIED_TYPE_APP

        entry.display_name = self._format_display_name(
            app_metadata.get('name'))
        entry.description = app_metadata.get('description')

        entry.linked_resource = f'{self.__site_url}' \
                                f'/sense/app/{app_metadata.get("id")}'

        created_datetime = datetime.strptime(
            app_metadata.get('createdDate'),
            self.__INCOMING_TIMESTAMP_UTC_FORMAT)
        create_timestamp = timestamp_pb2.Timestamp()
        create_timestamp.FromDatetime(created_datetime)
        entry.source_system_timestamps.create_time = create_timestamp

        modified_date = app_metadata.get('modifiedDate')
        resolved_modified_date = modified_date or app_metadata.get(
            'createdDate')
        modified_datetime = datetime.strptime(
            resolved_modified_date, self.__INCOMING_TIMESTAMP_UTC_FORMAT)
        update_timestamp = timestamp_pb2.Timestamp()
        update_timestamp.FromDatetime(modified_datetime)
        entry.source_system_timestamps.update_time = update_timestamp

        return generated_id, entry

    def make_entry_for_custom_property_definition(
            self, custom_property_def_metadata):

        entry = datacatalog.Entry()

        generated_id = self.__format_id(
            constants.ENTRY_ID_PART_CUSTOM_PROPERTY_DEFINITION,
            custom_property_def_metadata.get('id'))
        entry.name = datacatalog.DataCatalogClient.entry_path(
            self.__project_id, self.__location_id, self.__entry_group_id,
            generated_id)

        entry.user_specified_system = self.__user_specified_system
        entry.user_specified_type = \
            constants.USER_SPECIFIED_TYPE_CUSTOM_PROPERTY_DEFINITION

        entry.display_name = self._format_display_name(
            custom_property_def_metadata.get('name'))
        entry.description = custom_property_def_metadata.get('description')

        # The linked_resource field is not fulfilled because there is no way to
        # jump directly to an 'edit' page in the QlikView Management Console
        # (QMC). The the ID wee see in the URL of the Custom Property
        # Definition edit page is generated at the client side as a wrapper
        # around the object. The reason for this is: if someone select a bunch
        # of things in the QMC, it can't pick one, or have a list, so it
        # generates a new 'synthetic' key for the edit page.
        # -- from the Qlik Analytics Platform Architecture Team

        created_datetime = datetime.strptime(
            custom_property_def_metadata.get('createdDate'),
            self.__INCOMING_TIMESTAMP_UTC_FORMAT)
        create_timestamp = timestamp_pb2.Timestamp()
        create_timestamp.FromDatetime(created_datetime)
        entry.source_system_timestamps.create_time = create_timestamp

        modified_date = custom_property_def_metadata.get('modifiedDate')
        resolved_modified_date = \
            modified_date or custom_property_def_metadata.get('createdDate')
        modified_datetime = datetime.strptime(
            resolved_modified_date, self.__INCOMING_TIMESTAMP_UTC_FORMAT)
        update_timestamp = timestamp_pb2.Timestamp()
        update_timestamp.FromDatetime(modified_datetime)
        entry.source_system_timestamps.update_time = update_timestamp

        return generated_id, entry

    def make_entry_for_dimension(self, dimension_metadata):
        entry = datacatalog.Entry()

        app_metadata = dimension_metadata.get('app')

        # The Dimension ID is usually a 7 letters string, so the App ID is
        # prepended to prevent overlapping.
        generated_id = self.__format_id(
            constants.ENTRY_ID_PART_DIMENSION, f'{app_metadata.get("id")}'
            f'_{dimension_metadata.get("qInfo").get("qId")}')
        entry.name = datacatalog.DataCatalogClient.entry_path(
            self.__project_id, self.__location_id, self.__entry_group_id,
            generated_id)

        entry.user_specified_system = self.__user_specified_system
        entry.user_specified_type = constants.USER_SPECIFIED_TYPE_DIMENSION

        q_meta_def = dimension_metadata.get('qMetaDef')

        entry.display_name = self._format_display_name(q_meta_def.get('title'))
        entry.description = q_meta_def.get('description')

        # The linked_resource field is not fulfilled because there is no way to
        # jump directly to a Dimension 'edit' page in Qlik Sense.

        # The create_time and update_time fields are not fulfilled because
        # there is no such info in the Dimension metadata.

        return generated_id, entry

    def make_entry_for_measure(self, measure_metadata):
        entry = datacatalog.Entry()

        app_metadata = measure_metadata.get('app')

        # The Measure ID is usually a 7 letters string, so the App ID is
        # prepended to prevent overlapping.
        generated_id = self.__format_id(
            constants.ENTRY_ID_PART_MEASURE, f'{app_metadata.get("id")}'
            f'_{measure_metadata.get("qInfo").get("qId")}')
        entry.name = datacatalog.DataCatalogClient.entry_path(
            self.__project_id, self.__location_id, self.__entry_group_id,
            generated_id)

        entry.user_specified_system = self.__user_specified_system
        entry.user_specified_type = constants.USER_SPECIFIED_TYPE_MEASURE

        q_meta_def = measure_metadata.get('qMetaDef')

        entry.display_name = self._format_display_name(q_meta_def.get('title'))
        entry.description = q_meta_def.get('description')

        # The linked_resource field is not fulfilled because there is no way to
        # jump directly to a Measure 'edit' page in Qlik Sense.

        # The create_time and update_time fields are not fulfilled because
        # there is no such info in the Measure metadata.

        return generated_id, entry

    def make_entry_for_sheet(self, sheet_metadata):
        entry = datacatalog.Entry()

        sheet_id = sheet_metadata.get('qInfo').get('qId')
        generated_id = self.__format_id(constants.ENTRY_ID_PART_SHEET,
                                        sheet_id)
        entry.name = datacatalog.DataCatalogClient.entry_path(
            self.__project_id, self.__location_id, self.__entry_group_id,
            generated_id)

        entry.user_specified_system = self.__user_specified_system
        entry.user_specified_type = constants.USER_SPECIFIED_TYPE_SHEET

        q_meta = sheet_metadata.get('qMeta')
        entry.display_name = self._format_display_name(q_meta.get('title'))
        entry.description = q_meta.get('description')

        entry.linked_resource = f'{self.__site_url}' \
                                f'/sense/app/' \
                                f'{sheet_metadata.get("app").get("id")}' \
                                f'/sheet/{sheet_id}'

        created_datetime = datetime.strptime(
            q_meta.get('createdDate'), self.__INCOMING_TIMESTAMP_UTC_FORMAT)
        create_timestamp = timestamp_pb2.Timestamp()
        create_timestamp.FromDatetime(created_datetime)
        entry.source_system_timestamps.create_time = create_timestamp

        modified_date = q_meta.get('modifiedDate')
        resolved_modified_date = modified_date or q_meta.get('createdDate')
        modified_datetime = datetime.strptime(
            resolved_modified_date, self.__INCOMING_TIMESTAMP_UTC_FORMAT)
        update_timestamp = timestamp_pb2.Timestamp()
        update_timestamp.FromDatetime(modified_datetime)
        entry.source_system_timestamps.update_time = update_timestamp

        return generated_id, entry

    def make_entry_for_stream(self, stream_metadata):
        entry = datacatalog.Entry()

        generated_id = self.__format_id(constants.ENTRY_ID_PART_STREAM,
                                        stream_metadata.get('id'))
        entry.name = datacatalog.DataCatalogClient.entry_path(
            self.__project_id, self.__location_id, self.__entry_group_id,
            generated_id)

        entry.user_specified_system = self.__user_specified_system
        entry.user_specified_type = constants.USER_SPECIFIED_TYPE_STREAM

        entry.display_name = self._format_display_name(
            stream_metadata.get('name'))

        entry.linked_resource = f'{self.__site_url}' \
                                f'/hub/stream/{stream_metadata.get("id")}'

        created_datetime = datetime.strptime(
            stream_metadata.get('createdDate'),
            self.__INCOMING_TIMESTAMP_UTC_FORMAT)
        create_timestamp = timestamp_pb2.Timestamp()
        create_timestamp.FromDatetime(created_datetime)
        entry.source_system_timestamps.create_time = create_timestamp

        modified_date = stream_metadata.get('modifiedDate')
        resolved_modified_date = modified_date or stream_metadata.get(
            'createdDate')
        modified_datetime = datetime.strptime(
            resolved_modified_date, self.__INCOMING_TIMESTAMP_UTC_FORMAT)
        update_timestamp = timestamp_pb2.Timestamp()
        update_timestamp.FromDatetime(modified_datetime)
        entry.source_system_timestamps.update_time = update_timestamp

        return generated_id, entry

    def make_entry_for_visualization(self, visualization_metadata):
        entry = datacatalog.Entry()

        viz_id = visualization_metadata.get('qInfo').get('qId')
        generated_id = self.__format_id(constants.ENTRY_ID_PART_VISUALIZATION,
                                        viz_id)
        entry.name = datacatalog.DataCatalogClient.entry_path(
            self.__project_id, self.__location_id, self.__entry_group_id,
            generated_id)

        entry.user_specified_system = self.__user_specified_system
        entry.user_specified_type = constants.USER_SPECIFIED_TYPE_VISUALIZATION

        q_meta_def = visualization_metadata.get('qMetaDef')

        entry.display_name = self._format_display_name(q_meta_def.get('title'))
        entry.description = q_meta_def.get('description')

        # The linked_resource field is not fulfilled because Data Catalog
        # currently does not accept ``?`` and ``=`` in the field value.
        # The below statements can be uncommented once
        # https://issuetracker.google.com/issues/176912978
        # has been fixed.
        #
        # app_id = visualization_metadata.get('app').get('id')
        # entry.linked_resource = f'{self.__site_url}/sense/single' \
        #                         f'?appid={app_id}' \
        #                         f'&obj={viz_id}'

        # The create_time and update_time fields are not fulfilled because
        # there is no such info in the Visualization metadata.

        return generated_id, entry

    @classmethod
    def __format_id(cls, source_type_identifier, source_id):
        no_prefix_fmt_id = cls._format_id(
            f'{source_type_identifier}{source_id}')
        if len(no_prefix_fmt_id) > constants.NO_PREFIX_ENTRY_ID_MAX_LENGTH:
            no_prefix_fmt_id = \
                no_prefix_fmt_id[:constants.NO_PREFIX_ENTRY_ID_MAX_LENGTH]
        return f'{constants.ENTRY_ID_PREFIX}{no_prefix_fmt_id}'
